#! /bin/bash
# ex: set expandtab shiftwidth=4 softtabstop=4:
#
#
# For each study in the archive, query the number of study related objects on Insite
# check against number of local objects, log results

# Added support to check studies in a directory in data/dicom/styiuid directory format
# Added report-only (query only) mode
# Added a real option parser
# Added archive config file (read in ae titles, ports, ips, etc.)
# Added resume mode
# Added send only mode
# Added option to send obsoleted studies or not
# Add support to read device config from the device table
# Check archive table for obsolete studies (skip) ARCHST negative number = obsoleted

# To Do
# Add detailed verification mode (will require other scripts)

. ~/etc/setenv.sh
. $DICOM_VAR/pb-scp.cfg
#. remotearchive.cfg # no longer needed, just specify a hostid from the devices table

function listDevices() {
    echo "select * from DevServ;"|~/component/mysql/bin/mysql imagemedical
}


while getopts ":D:dhloq:r:s:S:T:t:vV" OPT ; do
    case $OPT in
        D) datadir="$OPTARG"
           noarchivedb="true"
           ;; 

        d) DLOG="1" ;;

        h) echo
           echo "Usage: `basename $0` -Ddvhn"
           echo "     -D <directory>        Full path to directory containing STYIUID directories"
           echo "     -d                    Detailed lists of failed objects/studies"
           echo "     -h                    Display help"
           echo "     -l                    List available hostids and supported services"
           echo "     -o                    Send obsoleted studies"
           echo "     -q <hostid>           Query number of study related objects from <hostid>"
           echo "     -r                    Full path to this script's output (for resuming starting position)"
           echo "     -s <hostid>           Send objects to <hostid>"
           echo "     -S <seconds>          Sleep <seconds> between studies"
           echo "     -t <seconds>          Set SEND timeout to <seconds>"
           echo "     -T <seconds>          Set RECEIVE timeout to <seconds>"
           echo "     -v                    Perform detailed verification"
           echo "     -V                    DICOM debug"
           exit 0
           ;;

	l) listDevices
           exit 0
           ;;

        o) sendobs=1 ;;

        q) query="$OPTARG" ;;

        r) resume="$OPTARG" ;;

        s) send="$OPTARG" ;;

        S) SLEEP="$OPTARG" ;;

        t) STO="$OPTARG" ;;

        T) RTO="$OPTARG" ;;

        v) analverify="1" ;;

        V) DEBUG="1" ;;

        *) echo "No such option $OPTARG"
           $0 -h
           exit 0
           ;;

    esac
done
[ -z "$query" -a -z "$send" ] && echo "$0 : Nothing to do (try the -q and/or -s option, or -h for full list)" && exit 0

if [ -n "$query" ]; then
    queryparams=$(echo "select HOST, PORT, AE from Target where ID='$query';" |\
    ~/component/mysql/bin/mysql -N -B imagemedical| \
    sed 's/\t/ /g') 
    set -- $queryparams
    QR_HOST=$1
    QR_PORT=$2
    QR_AE=$3
    echo "Query device: $QR_HOST, $QR_PORT, $QR_AE"
fi

if [ -n "$send" ]; then
    queryparams=$(echo "select HOST, PORT, AE from Target where ID='$send';" |\
    ~/component/mysql/bin/mysql -N -B imagemedical| \
    sed 's/\t/ /g') 
    set -- $queryparams
    CALL_AE=$3
    FWD_HOST=$1
    FWD_PORT=$2
    echo "Store device: $FWD_HOST, $FWD_PORT, $CALL_AE"
fi

SCU_MYAET=""
[ -z "$AE_TITLE" ] && AE_TITLE="$DICOM_AE_TITLE"
[ -z "$STORESCU_FLAG" ] && STORESCU_FLAG="-R"
[ -n "$DEBUG" ] && STORESCU_FLAG="$STORESCU_FLAG -v -d"
[ -n "$STO" ] && STORESCU_FLAG="$STORESCU_FLAG -st $STO"
[ -n "$RTO" ] && STORESCU_FLAG="$STORESCU_FLAG -rt $RTO"
[ -n "$AE_TITLE" ] && SCU_MYAET="-aet $AE_TITLE"
[ -z "$CALL_AE" ] && CALL_AE="--call some_Store_ae"
[ -z "$FWD_HOST" ] && FWD_HOST="10.10.1.150"
[ -z "$FWD_PORT" ] && FWD_PORT="10104"
[ -z "$QR_AE" ] && QR_AE="some_Query_ae"
[ -z "$QR_HOST" ] && QR_HOST="10.10.1.150"
[ -z "$QR_PORT" ] && QR_PORT="14444"

QR_ROOT="S"
SKIPTYPES="SRy" # Objects that Insite is known not to accept

#DEBUG="on"
WORKDIR=$(pwd)

# Error codes
SUCCESS=0
NEGATIVE_RESULT=1
INVALID_ARGS=2
FILE_NOT_FOUND=3
DATABASE_ERROR=4
DICOM_ERROR=5
ZEROLENGTH=6
SKIPPED=7
DUPLICATE=8

function hasHashedDicomdir() {

    [ -f /home/medsrv/component/utils/bin/getHashDirectory ] && return 0 || return 1
}


##############################################################################
#  _countObjects_()
#
#  Count the number of objects in a study
#  
#
#  Parameters:
#             studyuid
#             archive location
#
#  Returns:
#             a list of valid objects or 0 if none exist
#             positive number on failure
#
# expr "$string" : '\($substring\)'	Extract $substring* at beginning of $string
# element_count=0
# for element in ${myarray[*]};do
#   echo $element
#   [ $(expr match ${myarray[$element_count]} "def") -ne 0 ] && echo "Match" || echo "No Match! (${myarray[$element_count]}) != def"
#   (( element_count +=1 ));
# done
##############################################################################
_countObjects_() {

    local styiuid=$1
    local stydir=$2
    local objectarray
    local element
    local element_count=0
    local badtype
    local WHOLE
    local FILES
    local obsoleted_file

    if [ -d "$stydir" ]; then

        objectarray=($(ls -1 $stydir))

        for element in ${objectarray[@]}; do

            for badtype in $SKIPTYPES; do
                if [ $(expr match "$element" "$badtype") -ne 0 ]; then # Matches a badtype
                    unset objectarray[$element_count]
                    echo "$styiuid/$element : listed in badtypes : skipping" >> validateStudies.skipped
                fi
            done

            if [ -n "$FILES" ]; then
                for obsoleted_file in $FILES; do
                    if [ "$obsoleted_file" == "$element" ]; then
                        unset objectarray[$element_count]
                        echo "$styiuid/$element : object is obsoleted : skipping" >> validateStudies.skipped
                    fi
                done
            fi

            (( element_count += 1))

        done
                    
        echo ${objectarray[@]}

    else

        return

    fi

} # end of countObjects
    

##############################################################################
#  _setQuerycriteria_()
#
#  Set query criteria for Q/R
#  
#
#  Parameters:
#             none
#
#  Returns:
#             0 on success
#             positive number on failure
#
##############################################################################
_setQuerycriteria_() {

    echo "(0008,0052) CS [STUDY]
(0020,000d) UI [] # Study IUID
(0020,1208) IS [] # Number of Study related images
" > $WORKDIR/q.dmp

    $DCMTK_HOME/bin/dump2dcm $WORKDIR/q.dmp $WORKDIR/q.dcm >/dev/null 2>&1

    echo "(0008,0052) CS [STUDY]
" > $WORKDIR/s.dmp

    $DCMTK_HOME/bin/dump2dcm $WORKDIR/s.dmp $WORKDIR/s.dcm >/dev/null 2>&1

} # End of setQuerycriteria



##############################################################################
#  _queryStudy_()
#
#  Retrieve a study from external archive using Q/R
#  Acquire the number of study related objects that exist
#  
#
#  Parameters:
#             STYIUID - $1 - Study instance UID
#             SENTOBJS -$2 - Number of objects sent
#  Returns:
#             0 on success
#             positive number on failure
#
##############################################################################
_queryStudy_() {

    local styiuid=$1
    local sentobjs=$2

    [ -f q.res ] && rm -f q.res q.ls
    # get # of objects from the archive for this study UID
    #[ -n "$DEBUG" ] && echo -n "Requesting number of images for $styiuid:"
    echo -n "$styiuid:"

    $DICOM_HOME/bin/findscu -$QR_ROOT --aetitle $AE_TITLE --key 0020,000d="$styiuid" \
                                  --call $QR_AE $QR_HOST $QR_PORT $WORKDIR/q.dcm > q.res
    [ $? -ne 0 ] && return 3

    awk '
    $1=="(0020,000d)"  {s=$3;}
    $1=="(0020,1208)"  {if (s ~ "^\\[.*\\]$" ) { s= s "\t" $3; 
          gsub("\\[","",s); gsub("\\]","",s); print s; } }' q.res > q.ls

    # q.ls has styiuid and #objects - just need #objects
    if [ ! -f "q.res" ]; then
        echo "ERROR : $styiuid : Q/R failed?"
        return 3
    fi

    local queriedobjs=$(cat q.ls | cut -f 2)
    echo "_querystudy_: $sentobjs/$queriedobjs"

    if [ -z "$queriedobjs" ]; then
        echo "ERROR : $styiuid : not_on_target"
        if [ "$sentobjs" -gt 0 ]; then
            return 4
        fi
    elif [ "$queriedobjs" -ne "$sentobjs" ]; then
        echo "ERROR : $styiuid : $sentobjs local objects exist/queried $queriedobjs"
        if [ "$sentobjs" -gt "$queriedobjs" ]; then
            return 5
        else
            return 2
        fi
    else
        return 0
    fi

# Disable actual retrieve
#    echo "Retrieving $styiuid from archive:"
#    $DCMTK_HOME/bin/movescu -$QR_ROOT --move $AE_TITLE --aetitle $AE_TITLE \
#            --call $QR_AE $QR_HOST $QR_PORT --key 0020,000d="$styiuid" $WORKDIR/s.dcm
#    [ $? -ne 0 ] && return $DICOM_ERROR
#    return $SUCCESS

} # End of queryStudy




##############################################################################
#  _sendStudy_()
#
#  Parameters:
#             STYIUID - $1 - Study instance UID
#             archdir - $2 - Archive directory
#             objist  - $3 - Array of objects
#  Returns:
#             0 on success
#             positive number on failure
#
#[ -z "$AE_TITLE" ] && AE_TITLE="$DICOM_AE_TITLE"
#[ -z "$STORESCU_FLAG" ] && STORESCU_FLAG="-R"
#[ -n "$DEBUG" ] && STORESCU_FLAG="$STORESCU_FLAG -v -d"
#[ -n "$AE_TITLE" ] && SCU_MYAET="-aet $AE_TITLE"
#CALL_AE="--call indexStore-P07P"
#FWD_HOST="10.10.1.150"
#FWD_PORT="10104"
#QR_AE="indexQuery-P07P"
#QR_HOST="10.10.1.150"
#QR_PORT="14444"
#PBR_OPT="--pbr-file $DICOM_DATA/$Study/$PbRFile" 
# ${string##substring}
##############################################################################
_sendStudy_() {

    local styiuid=$1
    local stydir=$2
    local objlist
    objlist=($(echo "$3"))
    local objt=${#objlist[@]}
    local objs=0
    local objf=0
    local dobject
    local storeargs
    local PBR
    local PBR_OPT

    PBR=($(find $stydir -type f -mindepth 1 -maxdepth 1 -name "PbR.*"))
    if [ "${#PBR[@]}" -eq 1 ]; then
        status=$(${DICOM_HOME}/bin/dcmdump +P f215,1002 "${PBR}" | sed 's;^.*\[\(.*\)\].*$;\1;')
        if ! [ "${status#-}" = "$status" ]; then # It's an order
            echo "$styiuid : PbR is an order : skipping"
            return
        else
            PBR_OPT="--pbr-file $PBR"
        fi
    elif [ "${#PBR[@]}" -gt 1 ]; then
        echo "$styiuid : multiple PbR : skipping"
        return
    fi

    if [ -n "$DEBUG" ]; then
        echo "objlist: ${objlist[@]}"
        echo "objt: $objt"
        echo "styiuid: $styiuid"
        echo "stydir: $stydir"
    fi


    echo -n "Sending $styiuid : "

    for dobject in ${objlist[@]}; do

        if [ "$(expr "$dobject" : 'PbR')" -ne 0 ]; then # It's a PbR file
            storeargs="$STORESCU_FLAG $SCU_MYAET --call $CALL_AE $FWD_HOST $FWD_PORT $stydir/$dobject"
        else 
            storeargs="$STORESCU_FLAG $PBR_OPT $SCU_MYAET --call $CALL_AE $FWD_HOST $FWD_PORT $stydir/$dobject"
        fi
        if [ -n "$DEBUG" ]; then
            /home/medsrv/component/dicom/bin/storescu $storeargs >> storescu.debug.log
        else
            /home/medsrv/component/dicom/bin/storescu $storeargs > /dev/null 2>&1
        fi
        retval=$?
        if [ "$retval" -ne "0" ]; then
            (( objf += 1 ))
            if [ -n "$DLOG" ]; then
                echo "$styiuid/$dobject : failed sctorescu " >> storescu.detailed.log
            fi
        else
            (( objs += 1 ))
            if [ -n "$DLOG" ]; then
                echo "$styiuid/$dobject : success"  >> storescu.detailed.log
            fi

        fi
    done

    if [ "$objf" -ne "0" ]; then
        echo "storescu failed:t/s/f $objt/$objs/$objf"
        echo $styiuid >> failedstorescu.styiuid.log
    else
        echo "success:t/s/f $objt/$objs/$objf"
    fi

} # End of sendStudy



##############################################################################
#  _makeStudyList_()
#
#  Parameters:
#             None
#  Returns:
#             0 on success
#  Creates a file containing all of the SIUIDs in the archive           
#
##############################################################################

_makeStudyList_() {

if [ -f archivedstudy.ls ]; then
    read -p "A study list exists, use it? " resp
    [ "$resp" == 'y' ] && return 0
fi

find $HOME/data/dicom.repository -mindepth 4 -maxdepth 5 -type d -name "[0-9]*.[0-9]*" | while read dirnm; 
do
     echo $(basename $dirnm) 
done > archivedstudy.ls
retval=$?
if [ $retval -ne 0 ]; then
    echo "Error creating archivedstudy.ls, aborting"
    exit 101
fi

} # End of _makeStudyList_


##################################################################################
##
## Main
##
##################################################################################

## Create list of studies to check.

_makeStudyList_

if [ -n "$resume" ]; then
    [ -f done.ls ] && rm done.ls
    # Strip out any line that doesn't have "success" in it,
    #+ grab the styiuid from the rest, and write it to a file
    cat $resume | sed -e '/success/!d' | awk '{ print $2}' | sed 's/://g' > done.ls
    # done.ls is a list of styiuid's that have completed successfully
    # You can skip studies sent during multiple sessions by appending
    #+the output of the two runs into one file, then using it as the resume file
fi 

[ -n "$query" ] && _setQuerycriteria_

while read styiuid; do

    unset archloc
    unset numobjects
    unset objectlist

    if [ -n "$resume" ]; then
        if grep $styiuid done.ls > /dev/null 2>&1 ; then
            echo "$styiuid : skipped (resume mode)"
            continue
        fi
    fi 

    stydir=`$HOME/component/repositoryhandler/scripts/locateStudy.sh -d $styiuid`
    objectlist=($(_countObjects_ $styiuid $stydir))
    numobjects=${#objectlist[@]}
    echo "$styiuid - numobjects = $numobjects"    
    if [ "$numobjects" -ne "0" ]; then
        querysuccess=3
        querytrycount=0
        while [ "$querysuccess" -eq "3" ]; do
            if [ -n "$query" ]; then
                (( querytrycount += 1 ))
                _queryStudy_ $styiuid $numobjects
                querysuccess=$?
            else
                querysuccess=4
            fi
            if [ "$querysuccess" -gt "3" -a -n "$send" ]; then
                unset objectlist_arg # can't pass an array as an argument
                objectlist_arg=$(echo ${objectlist[@]})
                _sendStudy_ $styiuid $stydir "$objectlist_arg"
                [ -n "$SLEEP" ] && sleep $SLEEP
            fi
            if [ "$querytrycount" -eq "20" ]; then
                echo "Too many query retries - Target appears down. Exiting now"
                exit 100
            fi
        done 
    else
        echo "ERROR: $styiuid : nonexistant (in $archloc), obsoleted, or 0 objects"
    fi
done < archivedstudy.ls

